'' ********************************************
'' *               LockPool                   *
'' ********************************************

'' Author(s): Chad George
'' History: v0.1 - 25 July 2007, Original (C. George)

'' This object implements a generic pool of bytes,words or longs intented
'' to be shared among various objects throughout a project. If different
'' pools are needed then the filenames should be different otherwise
'' all instances will share the same data space. Each unqique instance
'' only needs to be initialized once.

'' All calls to public interface methods will be wrapped by a lock operation
'' in order to make the pool a "thread safe" mutually exclusive
'' resource. All methods block if needed and can be considered atomic.

'' The resource can be explicitly locked by the GetLock and ReleaseLock
'' methods. All calls between these two will have exclusive access to
'' the pool resource, but other cogs will block until the ReleaseLock is
'' called.

'' Built-in methods will only release a lock if they were responsible for
'' setting it. If the lock was already set they will leave it set on exit.

'' The object requires that at least one global semaphore is available but
'' does no explicit error checking.

'' The pool provides a form of dynamic memory allocation where arbitrary
'' contiguous blocks of the pool can be reserved. When the memory is
'' no longer required it is returned to the pool for general use.
  
PUB Initialize(pool_addr, pool_size, holetable_size, block_size)
'' Initializes (or resets) the shared resource and obtains the global
'' semphore resource. Only needs to be called once per instance.

  Destroy

  result := SemID := LockNew              
  
  if result <> -1
    PoolAddr := pool_addr
    PoolSize := pool_size
    FirstHole := -1

    InitHoleTable(holetable_size, block_size)

    Empty

                         
     
PUB Destroy
'' Releases the global resources

  if SemID <> -1
    GetLock
    ReleaseLock
    LockRet(SemID)

  SemID := -1
  LockID := -1
  
PRI InitHoleTable(table_size, block_size) | sz 
'' Initializes the hole table to be at the start of the memory pool

  
  HoleTableAddr := PoolAddr
  HoleTableSize := table_size
  BlockSize := block_size

  sz := ((HoleTableSize * SlotSize) / BlockSize) + 1                
  HoleTableBlocks := sz


  
PRI Empty | unlock, ptr, t1
   unlock := GetLock
   
  ' Clear the entire hole table
  ptr := HoleTableAddr

  
  repeat HoleTableSize
    word[ptr + StartAddress] := 0  
    word[ptr + HoleLength] := 0
    word[ptr + NextHole] := -1
    word[ptr + PrevHole] := -1
    ptr += SlotSize
    

  ' Add a single hole in the table 
  InsertHole(HoleTableBlocks, PoolSize - HoleTableBlocks)


  CollapseHoles


  
  if unlock
    ReleaseLock

{{ ----------------- Lock Management Routines ---------------------}}

PUB GetLock | once
'' Blocks in order to lock the shared resource unless the current cog already
'' has the lock. Returns true if the lock was checked out and
'' needs to be returned when finished

 result := False

 if LockID <> CogID
   repeat until not LockSet(SemID)                        'Wait until we can lock the ready queue
   LockID := CogID
   
 result := True
       
PUB ReleaseLock
'' Release the lock and its global semaphore
  LockID := -1
  LockClr(SemID)
   

PUB LockOwner
  return LockID


{{ ----------------- Manage the Hole Table Linked List ---------------------}}

PRI ChoseNextSlot | ptr
'' Finds the next available slot in the hole table to create a hole from.
'' If no slots are available it tries to double the size of the hole table.
'' Then it picks a new slot. The routine will fail if there is not
'' enough space to extend the hole table, or it is already at it's limit.

  ptr := HoleTableAddr

  repeat HoleTableSize
    if word[ptr + HoleLength] == 0
      return ptr
    ptr += SlotSize

  return -1

  
PRI InsertHole(address, length) | newptr, currptr, lastptr
'' Create an entry into the hole table from one of the available free slots 

  ' Try to find an empty slot in the hole table
  newptr := ChoseNextSlot

  if newptr == -1
    return False
    
  word[newptr + StartAddress] := address
  word[newptr + HoleLength] := length
  
  ' Find the insertion point
  currptr := FirstHole
  lastptr := -1
  
  repeat while (currptr <> -1) and word[currptr + StartAddress] < address
    lastptr := currptr
    currptr := word[currptr + NextHole]
    ~~currptr

  
  ' Insert the new node into the hole table list
  if lastptr <> -1
    word[lastptr + NextHole] := newptr
    
  'Fixup the current node
  if currptr <> -1
    word[currptr + PrevHole] := newptr
    
  if currptr == FirstHole
    FirstHole := newptr                                                      
    
  word[newptr + PrevHole] := lastptr
  word[newptr + NextHole] := currptr


  return True
  
PRI RemoveHole(holeptr) | prevptr, nextptr
'' Removes a hole from the hole table 

  ' Fixup links in the previos and next nodes
  nextptr := word[holeptr + NextHole]
  prevptr := word[holeptr + PrevHole]

  ~~nextptr
  ~~prevptr
  
  if nextptr <> -1
    word[nextptr + PrevHole] := prevptr

  if prevptr <> -1
    word[prevptr + NextHole] := nextptr

  if holeptr == FirstHole
    FirstHole := nextptr
    
  'Unlink the current node and clear the hole length
  'so the slot becomes available for use.
  word[holeptr + NextHole] := -1
  word[holeptr + PrevHole] := -1
  word[holeptr + HoleLength] := 0
  word[holeptr + StartAddress] := 0
  

PRI CollapseHoles | ptr, nextptr, lastptr, lastend, start, sz
'' Looks through the hole table for any two adjacent holes and collapses them
'' into a single hole. Also updates the MaxFree variable with the largest
'' hole found (or created) in the table.

    
    ptr := FirstHole
    lastptr := -1
    lastend := 0
    
    MaxFree := 0
    TotalFree := 0
    SlotsUsed := 0
    
    repeat while ptr <> -1
      start := word[ptr + StartAddress]
      sz := word[ptr + HoleLength]

      TotalFree += sz
      SlotsUsed += 1
      
      nextptr := word[ptr + NextHole]
      ~~nextptr
      
      if lastptr <> -1 and lastend == start
        word[lastptr + HoleLength] += sz
        if word[lastptr + HoleLength] > MaxFree
          MaxFree := word[lastptr + HoleLength]
            
        lastend += sz
        RemoveHole(ptr)
      else
        if sz > MaxFree
          MaxFree := sz

        lastend := start + sz
        lastptr := ptr

      ptr := nextptr
    
    
{{ ----------------- Public Interface Routines ---------------------}}

PUB RamToBlock(addr)
'' Converts a ram address pointer to a block address pointer
  result := (addr - PoolAddr) / BlockSize
  
PUB BlockToRam(addr)
'' Converts a block address pointer to a RAM address pointer
  result := PoolAddr + (addr * BlockSize)
    
PUB FindMaxFree | unlock, ptr, sz
'' Searches through the hole table to find the largest continuous block of memory

  unlock := GetLock


  ptr := FirstHole
  repeat while ptr <> -1
    sz := word[ptr + HoleLength]
    if sz > result
      result := sz
      
    ptr := word[ptr + NextHole]
    ~~ptr

  if unlock
    ReleaseLock

PUB FreePart(addr, size) | unlock
'' Returns a block of memory to the pool. Easier to use when only a portion of the
'' allocated memory will be returned to the pool. Fails if there are no remaining
'' free slots in the hole table and no room to expand the hole table.

  unlock := GetLock
   
  addr -= PoolAddr
  addr /= BlockSize
  
  result := InsertHole(addr, size)
  CollapseHoles

  if result
    TotalUsed -= size

  if unlock
    ReleaseLock

PUB Free(alloc_addr) | unlock, size, addr
'' Returns a block of memory to the pool. Expects the same format that the alloc
'' routines return. Fails if there are no remaining free slots in the hole table
'' and no room to expand the hole table.

  if alloc_addr == -1
    return False

  unlock := GetLock
  
  size := alloc_addr >> 16
  addr := alloc_addr & $FFFF

  addr -= PoolAddr
  addr /= BlockSize

  result := InsertHole(addr, size)

  if result
    TotalUsed -= size
    
  CollapseHoles

  if unlock
    ReleaseLock

PUB Alloc(size) | unlock, ptr, sz
'' Allocates a chuck of memory and returns a pointer to the start. Fails if
'' the the request size is too big. Returns the pointer in the lower 16 bits
'' and the size of the allocation in the upper 16 bits.


  unlock := GetLock

  ' Check if even the largest known hole is big enough
  result := -1
  
  if size =< MaxFree
    ' Find the first hole that is large enough to fill the request
    sz := 0

    ptr := FirstHole
    
    repeat while ptr <> -1
      sz := word[ptr + HoleLength]
      if sz => size
        quit
        
      ptr := word[ptr + NextHole]
      ~~ptr

    if ptr <> -1 and size =< word[ptr + HoleLength]
      ' Calculate the RAM address of the start of the allocated memory
      result := word[ptr + StartAddress] + word[ptr + HoleLength] - size
      result *= BlockSize
      result += PoolAddr
       
      result := (size << 16) | (result & $FFFF)
       
      ' Allocate the requested memory by making the hole smaller
      word[ptr + HoleLength] -= size
       
      ' Check if we allocated the entire hole
      if not word[ptr + HoleLength]
        RemoveHole(ptr)
        CollapseHoles

  if result <> -1
    TotalUsed += size

  if unlock
    ReleaseLock  


PUB AllocAny(size) | unlock
'' Tries to allocate the requested block of memory, but allocates the largest
'' block available if the requested size is too big.

  unlock := GetLock
  
  if size > MaxFree
    result := Alloc(MaxFree)
  else
    result := Alloc(size)

  if unlock
    ReleaseLock

PUB FreeMem | unlock
'' Returns the total number of blocks available.

  unlock := GetLock
  
  result := TotalFree
  if unlock
    ReleaseLock                        

PUB FreeSlots | unlock
'' Returns the number of empty slots available.
  unlock := GetLock
  
  result := HoleTableSize - SlotsUsed
  
  if unlock
    ReleaseLock                        

PUB MemStatus(ptr) | unlock
'' Takes a pointer to an array of 5 longs and fills it with the current
'' status of used and free memory.

  unlock := GetLock

  long[ptr]    := MaxFree
  long[ptr+=4] := TotalFree
  long[ptr+=4] := TotalUsed
  long[ptr+=4] := HoleTableSize
  long[ptr+=4] := SlotsUsed

  if unlock
    ReleaseLock                        
    
{{ ----------------- For Debugging Use Only ---------------------}}
PUB GetHoleTableAddr
  result := HoleTableAddr
  
PUB GetHoleTableSize
  result := HoleTableSize

PUB GetFirstHole
  result := FirstHole

PUB GetMaxFree
  result := MaxFree      

PUB GetTotalUsed
  result := TotalUsed

PUB GetTotalFree
  result := TotalFree
  
CON
  StartAddress  = 0
  HoleLength    = 2
  NextHole      = 4
  PrevHole      = 6
  SlotSize      = 8
       
DAT
SemID            word -1
LockID           word -1

PoolAddr         word 0                         'The address of the memory pool to be managed
PoolSize         word 0                         'The size of the pool in bytes

HoleTableAddr    word 0                         'The address of the hole table within the memory pool
HoleTableBlocks  word 0                         'Size of hole table in blocks

HoleTableSize    word 16                        'Total number of slots in the hole table
MaxHoleTableSize word 64                        'Maximum allowed size of the hole table                           
BlockSize        word 1

CurrHole         word 0                         'Address of the current position in the hole table (linked list)
FirstHole        word 0                         'Address of first hole in the hole table (linked list)

MaxFree          word 0                         'Always kept updated to the size of the largest hole
TotalUsed        word 0                         'Tracks the number of blocks that have been allocated
TotalFree        word 0                         'Tracks the total number of blocks available
SlotsUsed        word 0                         'Tracks the number of available slots.
                         