'' ********************************************
'' *              LockQueue                   *
'' ********************************************

'' Author(s): Chad George
'' History: v0.1 - 25 July 2007, Original (C. George)

'' This object implements a generic queue of bytes,words or longs intented
'' to be shared among various objects throughout a project. If different
'' queues are needed then the filenames should be different otherwise
'' all instances will share the same data space. Each unqique instance
'' only needs to be initialized once.

'' All calls to public interface methods will be wrapped by a lock operation
'' in order to make the queue a "thread safe" mutually exclusive
'' resource. All methods block if needed and can be considered atomic.
'' One expection is the IPeekHead and IPeekTail methods that never
'' block at the risk of a non-atomic operation.

'' The resource can be explicitly locked by the GetLock and ReleaseLock
'' methods. All calls between these two will have exclusive access to
'' the queue resource, but other cogs will block until the ReleaseLock is
'' called.

'' Built-in methods will only release a lock if they were responsible for
'' setting it. If the lock was already set they will leave it set on exit.

'' The object requires that at least one global semaphore is available but
'' does no explicit error checking.


CON
 
  EmptyQueue = -1           'sentinel returned for an empty queue
  BlockSize   = 4           'number of bytes per entry 1,2,or 4
  BlockShift  = 2           'shifting equiv of 1/DataSize or log2(DataSize) 

{{ ----------------- Initialization Routines ---------------------}}

PUB Initialize(queue_addr, queue_size)
'' Initializes (or resets) the shared resource and obtains the global
'' semphore resource. Only needs to be called once per instance.

  Destroy
    
  result := SemID := LockNew              

  if result <> -1
    QueueAddr := queue_addr
    QueueSize := queue_size

    QueueSpan := QueueSize * BlockSize    
    QueueEnd := QueueAddr + QueueSpan

    Empty
    
PUB Destroy
'' Releases the global resources
  if SemID <> -1
    GetLock
    ReleaseLock
    LockRet(SemID)

  SemID := -1
  LockID := -1  

    
PUB Empty | unlock
  unlock := GetLock

  QueueTail := QueueHead := QueueAddr

  if unlock
    ReleaseLock 

{{ ----------------- Lock Management Routines ---------------------}}

PUB GetLock
'' Blocks in order to lock the shared resource unless the current cog already
'' has the lock. Returns true if the lock was checked out and
'' needs to be returned when finished

  result := False
  if LockID <> CogID
    repeat until not LockSet(SemID)                       'Wait until we can lock the ready queue
    LockID := CogID
    result := True
       
PUB ReleaseLock
'' Release the lock and its global semaphore
  LockID := -1
  LockClr(SemID)

PUB LockOwner
  return LockID

{{ ----------------- Public Interface Routines ---------------------}}

PUB Count | unlock
'' Returns the number of items currently in the queue
  unlock := GetLock
  
  if QueueTail > QueueHead
    result := (QueueTail - QueueHead) >> BlockShift
  elseif QueueTail < QueueHead
    result := (QueueSpan - QueueHead + QueueTail) >> BlockShift

  if unlock
    ReleaseLock
  
PUB IPeekHead
  result := EmptyQueue

  if QueueHead <> QueueTail
    case BlockSize
      1:
        result := Byte[QueueHead]
        ~result
      2:
        result := Word[QueueHead]
        ~~result
      4: result := Long[QueueHead]

PUB IPeekTail
  result := EmptyQueue
  
  if QueueHead <> QueueTail
    case BlockSize
      1:
        result := Byte[QueueTail]
        ~result
      2:
        result := Word[QueueTail]
        ~~result
      4: result := Long[QueueTail]

PUB PeekHead | unlock
'' Retrieve the top value on the queue but don't remove it

  result := EmptyQueue
  unlock := GetLock
  
  if QueueHead <> QueueTail
    case BlockSize
      1:
        result := Byte[QueueHead]
        ~result
      2:
        result := Word[QueueHead]
        ~~result
      4: result := Long[QueueHead]

  if unlock
    ReleaseLock

PUB PeekTail | unlock , i
'' Retrieve the bottom value on the queue but don't remove it

  result := EmptyQueue
  unlock := GetLock

  i := QueueTail -= BlockSize
  if i < QueueAddr
      i += QueueSpan
        
  if QueueHead <> i
    case BlockSize
      1:
        result := Byte[i]
        ~result
      2: result := Word[i]
        ~~result
      4: result := Long[i]

  if unlock
    ReleaseLock

PUB Pop | unlock
'' Retrieve the value at the head of the queue

  result := EmptyQueue
  unlock := GetLock
   
  if QueueHead <> QueueTail
    case BlockSize
      1:
        result := Byte[QueueHead]
        ~result
      2:
        result := Word[QueueHead]
        ~~result
      4: result := Long[QueueHead]
  
    QueueHead += BlockSize
    if QueueHead => QueueEnd
      QueueHead -= QueueSpan
      
  if unlock
    ReleaseLock

PUB PopTail | unlock, i, pos
'' Retrieve the value at the tail of the queue

  result := EmptyQueue
  unlock := GetLock

  if QueueHead <> QueueTail

    i := QueueTail -= BlockSize
    if i < QueueAddr
      i += QueueSpan
      
    QueueTail := i
        
    case BlockSize
      1:
        result := Byte[QueueTail]
        ~result
      2:
        result := Word[QueueTail]
        ~~result
      4: result := Long[QueueTail]


      
  if unlock
    ReleaseLock
    
PUB Push(data) | unlock, i
'' Push the given value onto the back of the queue. Returns
'' the number of items added to the queue.

  result := 0
  unlock := GetLock

  'Check if there is room in the queue

  
  i := (QueueTail + BlockSize)                               
  if i => QueueEnd
    i -= QueueSpan
    
  if i <> QueueHead
    case BlockSize
      1: Byte[QueueTail] := data
      2: Word[QueueTail] := data
      4: Long[QueueTail] := data
      
    QueueTail := i     
    result := 1
  
  if unlock
    ReleaseLock

PUB PushHead(data) | unlock, i
'' Push the given value onto the front of the queue. Returns
'' the number of items added to the queue
  result := 0

  unlock := GetLock

  i := (QueueHead - BlockSize)
  if i < QueueAddr
    i += QueueSpan

  if i <> QueueTail
    QueueHead := i

    case BlockSize
      1: Byte[QueueHead] := data
      2: Word[QueueHead] := data
      4: Long[QueueHead] := data

    result := 1

  if unlock
    ReleaseLock

{{ ----------------- For Debugging Use Only ---------------------}}

PUB GetHead
  result := QueueHead
PUB GetTail
  result := QueueTail
PUB GetQueueAddr
  result := QueueAddr
PUB GetQueueSize
  result := QueueSize
  
DAT
SemID         word -1
LockID        word -1

QueueHead     word 0    'Pointer to start of queue
QueueTail     word 0    'Pointer to end of queue (next insertion point)

QueueAddr     word 0    'Lowest address in the queue
QueueSpan     word 0    'Span of the queue in bytes
QueueEnd      word 0    'Highest address in the queue

QueueSize     word 0
QueueSizeMask word 0